#!/usr/bin/env python3
#
#  properties.py
"""
Functions and classes to access properties of compounds in the PubChem database.
"""
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on PubChemPy https://github.com/mcs07/PubChemPy/blob/master/LICENSE
#  |  Copyright 2014 Matt Swain <m.swain@me.com>
#  |  Licensed under the MIT License
#  |
#  |  Permission is hereby granted, free of charge, to any person obtaining a copy
#  |  of this software and associated documentation files (the "Software"), to deal
#  |  in the Software without restriction, including without limitation the rights
#  |  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  |  copies of the Software, and to permit persons to whom the Software is
#  |  furnished to do so, subject to the following conditions:
#
#  |  The above copyright notice and this permission notice shall be included in
#  |  all copies or substantial portions of the Software.
#
#  |  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  |  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  |  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  |  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  |  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  |  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  |  THE SOFTWARE.
#

# stdlib
import warnings
from typing import Any, Callable, Dict, Iterable, List, NamedTuple, Optional, Sequence, Union

# 3rd party
from pandas import DataFrame  # type: ignore[import]

# this package
from chemistry_tools.formulae import Formula

# this package
from .enums import PubChemFormats, PubChemNamespace
from .pug_rest import do_rest_get
from .utils import _force_sequence_or_csv

__all__ = [
		"PropData",
		"valid_properties",
		"PROPERTY_MAP",
		"rest_get_properties_json",
		"rest_get_properties",
		"force_valid_properties",
		"get_properties",
		"get_property",
		"parse_properties",
		"PubChemProperty",
		]


class PropData(NamedTuple):
	"""
	Metadata about a property.
	"""

	#: The name of the property.
	name: str

	#: The description of the property.
	description: str

	#: The type of the property.
	type: Callable  # noqa: A003  # pylint: disable=redefined-builtin

	#: The Python attribute name of the property in a :class:`chemistry_tools.pubchem.compound.Compound`.
	attr_name: str


_properties: List[PropData] = [
		PropData("MolecularFormula", "Molecular formula.", Formula.from_string, "molecular_formula"),
		PropData(
				"MolecularWeight",
				"The molecular weight is the sum of all atomic weights of the constituent atoms "
				"in a compound, measured in g/mol. In the absence of explicit isotope labelling, "
				"averaged natural abundance is assumed. If an atom bears an explicit isotope label, "
				"100% isotopic purity is assumed at this location.",
				float,
				"molecular_weight"
				),
		PropData(
				"CanonicalSMILES",
				"Canonical SMILES (Simplified Molecular Input Line Entry System) string. "
				"It is a unique SMILES string of a compound, generated by a “canonicalization” algorithm.",
				str,
				"canonical_smiles"
				),
		PropData(
				"IsomericSMILES",
				"Isomeric SMILES string. It is a SMILES string with stereochemical and isotopic specifications.",
				str,
				"isomeric_smiles"
				),
		PropData(
				"InChI",
				"Standard IUPAC International Chemical Identifier (InChI). It does not allow for user "
				"selectable options in dealing with the stereochemistry and tautomer layers of the InChI string.",
				str,
				"inchi"
				),
		PropData(
				"InChIKey",
				"Hashed version of the full standard InChI, consisting of 27 characters.",
				str,
				"inchikey"
				),
		PropData(
				"IUPACName",
				"Chemical name systematically determined according to the IUPAC nomenclatures.",
				str,
				"iupac_name"
				),
		PropData(
				"XLogP",
				"Computationally generated octanol-water partition coefficient or distribution coefficient. "
				"XLogP is used as a measure of hydrophilicity or hydrophobicity of a molecule.",
				float,
				"xlogp"
				),
		PropData(
				"ExactMass",
				"The mass of the most likely isotopic composition for a single molecule, corresponding "
				"to the most intense ion/molecule peak in a mass spectrum.",
				float,
				"exact_mass"
				),
		PropData(
				"MonoisotopicMass",
				"The mass of a molecule, calculated using the mass of the most abundant isotope of each element.",
				float,
				"monoisotopic_mass"
				),
		PropData(
				"TPSA",
				"Topological polar surface area, computed by the algorithm described in the paper by Ertl et al.",
				float,
				"tpsa"
				),
		PropData(
				"Complexity",
				"The molecular complexity rating of a compound, computed using the Bertz/Hendrickson/Ihlenfeldt formula.",
				float,
				"complexity"
				),
		PropData("Charge", "The total (or net) charge of a molecule.", int, "charge"),
		PropData("HBondDonorCount", "Number of hydrogen-bond donors in the structure.", int, "h_bond_donor_count"),
		PropData(
				"HBondAcceptorCount",
				"Number of hydrogen-bond acceptors in the structure.",
				int,
				"h_bond_acceptor_count"
				),
		PropData("RotatableBondCount", "Number of rotatable bonds.", int, "rotatable_bond_count"),
		PropData("HeavyAtomCount", "Number of non-hydrogen atoms.", int, "heavy_atom_count"),
		PropData("IsotopeAtomCount", "Number of atoms with enriched isotope(s)", int, "isotope_atom_count"),
		PropData(
				"AtomStereoCount",
				"Total number of atoms with tetrahedral (sp3) stereo [e.g., (R)- or (S)-configuration]",
				int,
				"atom_stereo_count"
				),
		PropData(
				"DefinedAtomStereoCount",
				"Number of atoms with defined tetrahedral (sp3) stereo.",
				int,
				"defined_atom_stereo_count"
				),
		PropData(
				"UndefinedAtomStereoCount",
				"Number of atoms with undefined tetrahedral (sp3) stereo.",
				int,
				"undefined_atom_stereo_count"
				),
		PropData(
				"BondStereoCount",
				"Total number of bonds with planar (sp2) stereo [e.g., (E)- or (Z)-configuration].",
				int,
				"bond_stereo_count"
				),
		PropData(
				"DefinedBondStereoCount",
				"Number of atoms with defined planar (sp2) stereo.",
				int,
				"defined_bond_stereo_count"
				),
		PropData(
				"UndefinedBondStereoCount",
				"Number of atoms with undefined planar (sp2) stereo.",
				int,
				"undefined_bond_stereo_count"
				),
		PropData("CovalentUnitCount", "Number of covalently bound units.", int, "covalent_unit_count"),
		PropData(
				"Volume3D",
				"Analytic volume of the first diverse conformer (default conformer) for a compound.",
				str,
				"volume3d"
				),
		PropData(
				"XStericQuadrupole3D",
				"The x component of the quadrupole moment (Qx) of the first diverse conformer (default conformer) "
				"for a compound.",
				float,
				"volume_3d"
				),
		PropData(
				"YStericQuadrupole3D",
				"The y component of the quadrupole moment (Qy) of the first diverse conformer (default conformer) "
				"for a compound.",
				float,
				"x_steric_quadrupole_3d"
				),
		PropData(
				"ZStericQuadrupole3D",
				"The z component of the quadrupole moment (Qz) of the first diverse conformer (default conformer) "
				"for a compound.",
				float,
				"y_steric_quadrupole_3d"
				),
		PropData(
				"FeatureCount3D",
				"Total number of 3D features (the sum of FeatureAcceptorCount3D, FeatureDonorCount3D, "
				"FeatureAnionCount3D, FeatureCationCount3D, FeatureRingCount3D and FeatureHydrophobeCount3D)",
				int,
				"feature_count_3d"
				),
		PropData(
				"FeatureAcceptorCount3D",
				"Number of hydrogen-bond acceptors of a conformer.",
				int,
				"feature_acceptor_count_3d"
				),
		PropData(
				"FeatureDonorCount3D",
				"Number of hydrogen-bond donors of a conformer.",
				int,
				"feature_donor_count_3d"
				),
		PropData(
				"FeatureAnionCount3D",
				"Number of anionic centers (at pH 7) of a conformer.",
				int,
				"feature_anion_count_3d"
				),
		PropData(
				"FeatureCationCount3D",
				"Number of cationic centers (at pH 7) of a conformer.",
				int,
				"feature_cation_count_3d"
				),
		PropData("FeatureRingCount3D", "Number of rings of a conformer.", int, "feature_ring_count_3d"),
		PropData(
				"FeatureHydrophobeCount3D",
				"Number of hydrophobes of a conformer.",
				int,
				"feature_hydrophobe_count_3d"
				),
		PropData("ConformerModelRMSD3D", "Conformer sampling RMSD in Å.", float, "conformer_model_rmsd_3d"),
		PropData(
				"EffectiveRotorCount3D",
				"Total number of 3D features (the sum of FeatureAcceptorCount3D, FeatureDonorCount3D, FeatureAnionCount3D, "
				"FeatureCationCount3D, FeatureRingCount3D and FeatureHydrophobeCount3D)",
				int,
				"effective_rotor_count_3d"
				),
		PropData(
				"ConformerCount3D",
				"The number of conformers in the conformer model for a compound.",
				int,
				"conformer_count_3d"
				),
		PropData(
				"Fingerprint2D",
				"Base64-encoded PubChem Substructure Fingerprint of a molecule.",
				str,
				"fingerprint_2d"
				),
		]

#: Properties for PubChem REST API
valid_properties: Dict[str, Callable] = {prop.name: prop.type for prop in _properties}

#: Allows properties to optionally be specified as underscore_separated, consistent with Compound attributes
PROPERTY_MAP: Dict[str, str] = {prop.attr_name: prop.name for prop in _properties}


def rest_get_properties_json(
		identifier: Union[str, int, Sequence[Union[str, int]]],
		namespace: Union[str, PubChemNamespace] = PubChemNamespace.name,
		properties: Union[Sequence[str], str] = '',
		**kwargs
		) -> Dict:
	"""
	Returns the properties for the compound with the given identifier as a dictionary.

	:param identifier: Identifiers (e.g. name, CID) for the compound to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:param namespace: The type of identifier to look up. Valid values are in :class:`~.PubChemNamespace`.
	:param properties: The properties to retrieve for the compound.
		Can be either a comma-separated string or a list.
		See :ref:`the table at the start of this chapter <properties table>` for a list of valid properties.

	:param kwargs: Optional arguments that ``json.loads`` takes.

	:raises ValueError: If the response body does not contain valid JSON.

	:return: Parsed JSON data
	"""

	properties = force_valid_properties(properties)

	for prop in properties:
		if prop not in valid_properties:
			raise ValueError(f"Unknown property '{prop}'")

	return do_rest_get(namespace, identifier, domain=f"property/{','.join(properties)}").json(**kwargs)


def rest_get_properties(
		identifier: Union[str, int, Sequence[Union[str, int]]],
		namespace=PubChemNamespace.name,
		properties: Union[Sequence[str], str] = '',
		format_: Union[PubChemFormats, str] = PubChemFormats.CSV,
		) -> str:
	r"""
	Returns the properties for the compound with the given identifier in the desired format.

	:param identifier: Identifiers (e.g. name, CID) for the compound to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:param namespace: The type of identifier to look up. Valid values are in :class:`~.PubChemNamespace`
	:param properties: The properties to retrieve for the compound.
		Can be either a comma-separated string or a list.
		See :ref:`the table at the start of this chapter <properties table>` for a list of valid properties.

	:param format\_: The format to obtain the data in
	"""

	properties = force_valid_properties(properties)

	for prop in properties:
		if prop not in valid_properties:
			raise ValueError(f"Unknown property '{prop}'")

	return do_rest_get(namespace, identifier, domain=f"property/{','.join(properties)}", format_=format_).text


def force_valid_properties(properties: Union[str, Iterable[str]]) -> List[str]:
	"""
	Coerce ``properties`` into a list of strings and exclude any invalid properties,
	or raise a :exc:`ValueError` if that is not possible.

	:param properties:
	"""  # noqa: D400

	properties = _force_sequence_or_csv(properties, "properties")

	ordered_properties = []

	for prop in valid_properties:
		if prop in properties:
			properties.remove(prop)
			ordered_properties.append(prop)

	if properties:
		raise ValueError(f"Unknown properties '{', '.join(properties)}'")

	if not ordered_properties:
		raise ValueError(f"Please supply one or more properties")

	return ordered_properties


def get_properties(
		identifier: Union[str, int, Sequence[Union[str, int]]],
		properties: Union[Sequence[str], str] = '',
		namespace: Union[PubChemNamespace, str] = PubChemNamespace.name,
		as_dataframe: bool = False,
		) -> Union[List[Dict[str, Any]], DataFrame]:
	"""
	Returns the requested properties for the compound with the given identifier.
	As more than one compound may be identified the results are returned in a list.

	:param identifier: Identifiers (e.g. name, CID) for the compound to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.

	:param properties: The properties to retrieve for the compound.
		Can be either a comma-separated string or a list.
		See :ref:`the table at the start of this chapter <properties table>` for a list of valid properties.

	:param namespace: The type of identifier to look up. Valid values are in :class:`~.PubChemNamespace`.
	:param as_dataframe: Automatically extract the properties into a pandas :class:`~pandas.DataFrame`.

	:raises ValueError: If the response body does not contain valid JSON.
	:raises NotFoundError: If the compound with the requested identifier was not found in PubChem.

	:return: List of dictionaries mapping properties to values
	"""

	if isinstance(properties, str) and properties.lower() == "all":
		properties = list(valid_properties.keys())

	properties = force_valid_properties(properties)

	data = rest_get_properties_json(identifier, namespace, properties)

	results = []

	for compound in parse_properties(data):
		parsed_data = {"CID": compound["CID"]}

		for prop in properties:
			parsed_data[prop] = compound[prop]

		results.append(parsed_data)

	if as_dataframe:
		return DataFrame.from_records(results, index="CID")

	return results


def get_property(
		identifier: Union[str, int, Sequence[Union[str, int]]],
		property: str = '',  # noqa: A002  # pylint: disable=redefined-builtin
		namespace: Union[PubChemNamespace, str] = PubChemNamespace.name
		) -> Any:
	"""
	Returns the requested property for the compound with the given identifier.

	This convenience function only allows for a single property to be accessed at once,
	and for only a single compound. if you require multiple properties and/or properties for
	multiple compounds use :class:`chemistry_tools.pubchem.properties.get_properties`, which helps
	reduce the burden on the PubChem servers.

	:param identifier: Identifiers (e.g. name, CID) for the compound to look up.

	:param properties: The properties to retrieve for the compound.
		Can be either a comma-separated string or a list.
		See :ref:`the table at the start of this chapter <properties table>` for a list of valid properties.

	:param namespace: The type of identifier to look up. Valid values are in :class:`~.PubChemNamespace`.

	:raises ValueError: If the response body does not contain valid JSON.
	:raises NotFoundError: If the compound with the requested identifier was not found in PubChem.

	:return: The requested property. Type depends on the property requested.

	.. latex:clearpage::
	"""

	property_ = force_valid_properties(property)[0]

	data = rest_get_properties_json(identifier, namespace, [property_])

	return parse_properties(data)[0][property_]


def parse_properties(property_data: Dict) -> List[Dict]:
	"""
	Parse raw data from the ``property`` endpoint of the REST API.

	:param property_data:

	:return: A list of dictionaries mapping the properties to values for each compound
	"""

	compounds = {}
	fields = valid_properties

	for entry in property_data["PropertyTable"]["Properties"]:

		cid = entry["CID"]

		if cid not in compounds:
			compounds[cid] = {var: None for var in fields}
			compounds[cid]["CID"] = cid

		for var in fields:
			if var in entry:
				property_type = valid_properties[var]

				if property_type is int and isinstance(entry[var], float):
					warnings.warn(f"Loss of precision converting {var} from float to int.")

				if property_type is not Any:
					compounds[cid][var] = property_type(entry[var])
				else:
					compounds[cid][var] = entry[var]

	return list(compounds.values())


class __BasePubChemProperty(NamedTuple):
	label: str
	name: Optional[str]
	value: Any
	dtype: Callable
	source: Dict


class PubChemProperty(__BasePubChemProperty):
	"""
	Represents a property parsed from the full PubChem record.

	:param label: The label of the property.
	:param name: The name of the property.
	:param value: The property's value.
	:param dtype: The data type property's value.
	:param source: Dictionary of property sources.
	"""

	__slots__: List[str] = []

	def __new__(cls, label: str, name: Optional[str] = None, value=None, dtype=None, source=None):  # noqa: D102
		if source is None:
			source = {}

		if dtype:
			value = dtype(value)
		label = str(label)
		if name:
			name = str(name)

		return super().__new__(cls, label, name, value, dtype, source)


def string_list(val: Iterable[Any]) -> List[str]:
	"""
	Coerce an iterable of values to a list of strings.

	:param val:
	"""

	return [str(x) for x in val]


def _parse_record_property(prop: Dict) -> PubChemProperty:
	urn = prop["urn"]

	if "name" in urn:
		prop_name = urn["name"]
	else:
		prop_name = None

	dtype = urn["datatype"]

	prop_dtype: Callable

	if dtype == 7:
		prop_dtype = float
		val_key = "fval"
	elif dtype == 1:
		prop_dtype = str
		val_key = "sval"
	elif dtype == 16:
		prop_dtype = str
		val_key = "binary"
	elif dtype == 5:
		prop_dtype = int
		val_key = "ival"
	elif dtype == 2:
		prop_dtype = string_list
		val_key = "slist"
	else:
		raise ValueError(f"Unknown datatype '{dtype}' for property {prop_name} {urn['label']}")

	value = prop["value"][val_key]

	prop_source = {k: val for k, val in urn.items() if k not in {"datatype", "label", "name"}}

	return PubChemProperty(label=urn["label"], name=prop_name, value=value, dtype=prop_dtype, source=prop_source)



#!/usr/bin/env python3
#
#  utils.py
"""
General utility functions.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Also based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |

# stdlib
import re
from functools import lru_cache
from typing import Dict, Iterator, List, Sequence, Tuple

# this package
from chemistry_tools.elements import ELEMENTS

__all__ = [
		"GROUPS",
		"split_isotope",
		"hill_order",
		]

#: Common chemical groups
GROUPS: Dict[str, str] = {
		"Abu": "C4H7NO",
		"Acet": "C2H3O",
		"Acm": "C3H6NO",
		"Adao": "C10H15O",
		"Aib": "C4H7NO",
		"Ala": "C3H5NO",
		"Arg": "C6H12N4O",
		"Argp": "C6H11N4O",
		"Asn": "C4H6N2O2",
		"Asnp": "C4H5N2O2",
		"Asp": "C4H5NO3",
		"Aspp": "C4H4NO3",
		"Asu": "C8H13NO3",
		"Asup": "C8H12NO3",
		"Boc": "C5H9O2",
		"Bom": "C8H9O",
		"Bpy": "C10H8N2",  # Bipyridine
		"Brz": "C8H6BrO2",
		"Bu": "C4H9",
		"Bum": "C5H11O",
		"Bz": "C7H5O",
		"Bzl": "C7H7",
		"Bzlo": "C7H7O",
		"Cha": "C9H15NO",
		"Chxo": "C6H11O",
		"Cit": "C6H11N3O2",
		"Citp": "C6H10N3O2",
		"Clz": "C8H6ClO2",
		"Cp": "C5H5",
		"Cy": "C6H11",
		"Cys": "C3H5NOS",
		"Cysp": "C3H4NOS",
		"Dde": "C10H13O2",
		"Dnp": "C6H3N2O4",
		"Et": "C2H5",
		"Fmoc": "C15H11O2",
		"For": "CHO",
		"Gln": "C5H8N2O2",
		"Glnp": "C5H7N2O2",
		"Glp": "C5H5NO2",
		"Glu": "C5H7NO3",
		"Glup": "C5H6NO3",
		"Gly": "C2H3NO",
		"Hci": "C7H13N3O2",
		"Hcip": "C7H12N3O2",
		"His": "C6H7N3O",
		"Hisp": "C6H6N3O",
		"Hser": "C4H7NO2",
		"Hserp": "C4H6NO2",
		"Hx": "C6H11",
		"Hyp": "C5H7NO2",
		"Hypp": "C5H6NO2",
		"Ile": "C6H11NO",
		"Ivdde": "C14H21O2",
		"Leu": "C6H11NO",
		"Lys": "C6H12N2O",
		"Lysp": "C6H11N2O",
		"Mbh": "C15H15O2",
		"Me": "CH3",
		"Mebzl": "C8H9",
		"Meobzl": "C8H9O",
		"Met": "C5H9NOS",
		"Mmt": "C20H17O",
		"Mtc": "C14H19O3S",
		"Mtr": "C10H13O3S",
		"Mts": "C9H11O2S",
		"Mtt": "C20H17",
		"Nle": "C6H11NO",
		"Npys": "C5H3N2O2S",
		"Nva": "C5H9NO",
		"Odmab": "C20H26NO3",
		"Orn": "C5H10N2O",
		"Ornp": "C5H9N2O",
		"Pbf": "C13H17O3S",
		"Pen": "C5H9NOS",
		"Penp": "C5H8NOS",
		"Ph": "C6H5",
		"Phe": "C9H9NO",
		"Phepcl": "C9H8ClNO",
		"Phg": "C8H7NO",
		"Pmc": "C14H19O3S",
		"Ppa": "C8H7O2",
		"Pro": "C5H7NO",
		"Prop": "C3H7",
		"Py": "C5H5N",
		"Pyr": "C5H5NO2",
		"Sar": "C3H5NO",
		"Ser": "C3H5NO2",
		"Serp": "C3H4NO2",
		"Sta": "C8H15NO2",
		"Stap": "C8H14NO2",
		"Tacm": "C6H12NO",
		"Tbdms": "C6H15Si",
		"Tbu": "C4H9",
		"Tbuo": "C4H9O",
		"Tbuthio": "C4H9S",
		"Tfa": "C2F3O",
		"Thi": "C7H7NOS",
		"Thr": "C4H7NO2",
		"Thrp": "C4H6NO2",
		"Tips": "C9H21Si",
		"Tms": "C3H9Si",
		"Tos": "C7H7O2S",
		"Trp": "C11H10N2O",
		"Trpp": "C11H9N2O",
		"Trt": "C19H15",
		"Tyr": "C9H9NO2",
		"Tyrp": "C9H8NO2",
		"Val": "C5H9NO",
		"Valoh": "C5H9NO2",
		"Valohp": "C5H8NO2",
		"Xan": "C13H9O",
		}

_isotope_regex_1 = re.compile(r"^([A-z]+)(\[\d*])$")
_isotope_regex_2 = re.compile(r"^(\[[A-z]+)(\d*])$")
_isotope_regex_3 = re.compile(r"^(\[\d*)([A-z]+])$")
_iso_bracket_regex = re.compile(r"^(\[)(\d+)(])$")
_hill_carbon_re = re.compile(r"(C(?:\[[0-9]+])?|\[[0-9]+C])")
_hill_hydrogen_re = re.compile(r"(H(?:\[[0-9]+])?|\[[0-9]+H])")


@lru_cache()
def split_isotope(string: str) -> Tuple[str, int]:
	"""
	Returns the symbol and mass number for the isotope represented by ``string``.

	Valid isotopes include ``'[C12]'``, ``'C[12]'`` and ``'[12C]'``.

	:param string:

	:return: Tuple representing the element and the isotope number.
	"""

	isotope = '0'

	iso_re_1 = _isotope_regex_1.findall(string)
	iso_re_2 = _isotope_regex_2.findall(string)
	iso_re_3 = _isotope_regex_3.findall(string)

	if iso_re_1:
		elem, isotope = iso_re_1[0]
		isotope = _iso_bracket_regex.findall(isotope)[0][1]
	elif iso_re_2:
		elem, isotope = iso_re_2[0]
		elem = elem.lstrip('[')
		isotope = isotope.rstrip(']')
	elif iso_re_3:
		isotope, elem = iso_re_3[0]
		isotope = isotope.lstrip('[')
		elem = elem.rstrip(']')
	else:
		elem = string

	if elem not in ELEMENTS:
		raise ValueError(f"Unknown chemical element with symbol {elem}")

	return ELEMENTS[elem].symbol, int(isotope)


def hill_order(symbols: Sequence[str]) -> Iterator[str]:
	"""
	Returns an iterator over the given element symbols in order of Hill notation.

	:bold-title:`Example:`

	.. code-block:: python

		>>> for i in hill_order("H", "C[12]", "O"): print(i, end='')
		CHO
	"""

	symbols_list: List[str] = list(set(symbols))

	carbon_isotopes = list(filter(_hill_carbon_re.findall, symbols_list))
	print(carbon_isotopes)

	if carbon_isotopes:
		for isotope in sorted(carbon_isotopes):
			symbols_list.remove(isotope)
			yield isotope

		hydrogen_isotopes = list(filter(_hill_hydrogen_re.findall, symbols_list))
		for isotope in sorted(hydrogen_isotopes):
			symbols_list.remove(isotope)
			yield isotope

	yield from sorted(symbols_list)



#!/usr/bin/env python3
#
#  chalcogens.py
r"""
Group 16: Chalcogens in the Periodic Table.

.. raw:: latex

	\begin{multicols}{2}

.. data:: O

	:class:`~chemistry_tools.elements.classes.Element` representing Oxygen

.. data:: S

	:class:`~chemistry_tools.elements.classes.Element` representing Sulfur

.. data:: Se

	:class:`~chemistry_tools.elements.classes.Element` representing Selenium

.. data:: Te

	:class:`~chemistry_tools.elements.classes.Element` representing Tellurium

.. data:: Po

	:class:`~chemistry_tools.elements.classes.Element` representing Polonium

.. data:: Lv

	:class:`~chemistry_tools.elements.classes.Element` representing Livermorium

.. raw:: latex

	\end{multicols}
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Isotope data from http://www.nist.gov/pml/data/comp.cfm .

# this package
from .classes import Element

__all__ = ('O', 'S', "Se", "Te", "Po", "Lv")

# Isotope 0 Key:
# mass of the most abundant isotope and 1.0 abundance.

O = Element(
		8,
		'O',
		"Oxygen",
		group=16,
		period=2,
		block='p',
		series=1,
		mass=15.999405,
		eleneg=3.44,
		eleaffin=1.461112,
		covrad=0.73,
		atmrad=0.65,
		vdwrad=1.52,
		tboil=90.188,
		tmelt=54.8,
		density=1.33,
		eleconfig="[He] 2s2 2p4",
		oxistates="-2*, -1",
		ionenergy=(
				13.6181,
				35.116,
				54.934,
				77.412,
				113.896,
				138.116,
				739.315,
				871.387,
				),
		isotopes={
				12: (12.034405, 0.0),
				13: (13.024812, 0.0),
				14: (14.00859625, 0.0),
				15: (15.0030656, 0.0),
				16: (15.99491461956, 0.99757),
				17: (16.9991317, 0.00038),
				18: (17.999161, 0.00205),
				19: (19.00358, 0.0),
				20: (20.0040767, 0.0),
				21: (21.008656, 0.0),
				22: (22.00997, 0.0),
				23: (23.01569, 0.0),
				24: (24.02047, 0.0),
				25: (25.02946, 0.0),
				26: (26.03834, 0.0),
				27: (27.04826, 0.0),
				28: (28.05781, 0.0),
				},
		description=(
				"A colourless, odourless gaseous element belonging to group 16 of "
				"the periodic table. It is the most abundant element present in the "
				"Earth's crust. It also makes up 20.8% of the Earth's atmosphere. "
				"For industrial purposes, it is separated from liquid air by "
				"fractional distillation. It is used in high temperature welding, "
				"and in breathing. It commonly comes in the form of Oxygen, but is "
				"found as Ozone in the upper atmosphere. It was discovered by "
				"Priestley in 1774."
				)
		)

S = Element(
		16,
		'S',
		"Sulfur",
		group=16,
		period=3,
		block='p',
		series=1,
		mass=32.0648,
		eleneg=2.58,
		eleaffin=2.0771029,
		covrad=1.02,
		atmrad=1.09,
		vdwrad=1.8,
		tboil=717.82,
		tmelt=392.2,
		density=2.06,
		eleconfig="[Ne] 3s2 3p4",
		oxistates="6*, 4, 2, -2",
		ionenergy=(
				10.36,
				23.33,
				34.83,
				47.3,
				72.68,
				88.049,
				280.93,
				328.23,
				379.1,
				447.09,
				504.78,
				564.65,
				651.63,
				707.14,
				3223.836,
				3494.099,
				),
		isotopes={
				26: (26.02788, 0.0),
				27: (27.01883, 0.0),
				28: (28.00437, 0.0),
				29: (28.99661, 0.0),
				30: (29.984903, 0.0),
				31: (30.9795547, 0.0),
				32: (31.972071, 0.9499),
				33: (32.97145876, 0.0075),
				34: (33.9678669, 0.0425),
				35: (34.96903216, 0.0),
				36: (35.96708076, 0.0001),
				37: (36.97112557, 0.0),
				38: (37.971163, 0.0),
				39: (38.97513, 0.0),
				40: (39.97545, 0.0),
				41: (40.97958, 0.0),
				42: (41.98102, 0.0),
				43: (42.98715, 0.0),
				44: (43.99021, 0.0),
				45: (44.99651, 0.0),
				46: (46.00075, 0.0),
				47: (47.00859, 0.0),
				48: (48.01417, 0.0),
				49: (49.02362, 0.0),
				},
		description=(
				"Yellow, nonmetallic element belonging to group 16 of the periodic "
				"table. It is an essential element in living organisms, needed in "
				"the amino acids cysteine and methionine, and hence in many "
				"proteins. Absorbed by plants from the soil as sulphate ion."
				)
		)

Se = Element(
		34,
		"Se",
		"Selenium",
		group=16,
		period=4,
		block='p',
		series=1,
		mass=78.971,
		eleneg=2.55,
		eleaffin=2.02067,
		covrad=1.16,
		atmrad=1.22,
		vdwrad=1.9,
		tboil=958.0,
		tmelt=494.0,
		density=4.82,
		eleconfig="[Ar] 3d10 4s2 4p4",
		oxistates="6, 4*, -2",
		ionenergy=(
				9.7524,
				21.9,
				30.82,
				42.944,
				68.3,
				81.7,
				155.4,
				),
		isotopes={
				65: (64.96466, 0.0),
				66: (65.95521, 0.0),
				67: (66.95009, 0.0),
				68: (67.9418, 0.0),
				69: (68.93956, 0.0),
				70: (69.93339, 0.0),
				71: (70.93224, 0.0),
				72: (71.927112, 0.0),
				73: (72.926765, 0.0),
				74: (73.9224764, 0.0089),
				75: (74.9225234, 0.0),
				76: (75.9192136, 0.0937),
				77: (76.919914, 0.0763),
				78: (77.9173091, 0.2377),
				79: (78.9184991, 0.0),
				80: (79.9165213, 0.4961),
				81: (80.9179925, 0.0),
				82: (81.9166994, 0.0873),
				83: (82.919118, 0.0),
				84: (83.918462, 0.0),
				85: (84.92225, 0.0),
				86: (85.924272, 0.0),
				87: (86.92852, 0.0),
				88: (87.93142, 0.0),
				89: (88.93645, 0.0),
				90: (89.93996, 0.0),
				91: (90.94596, 0.0),
				92: (91.94992, 0.0),
				93: (92.95629, 0.0),
				94: (93.96049, 0.0),
				},
		description=(
				"Metalloid element, belongs to group 16 of the periodic table. "
				"Multiple allotropic forms exist. Chemically resembles sulphur. "
				"Discovered in 1817 by Jons J. Berzelius."
				)
		)

Te = Element(
		52,
		"Te",
		"Tellurium",
		group=16,
		period=5,
		block='p',
		series=5,
		mass=127.6,
		eleneg=2.1,
		eleaffin=1.970875,
		covrad=1.36,
		atmrad=1.42,
		vdwrad=2.06,
		tboil=1261.0,
		tmelt=722.72,
		density=6.25,
		eleconfig="[Kr] 4d10 5s2 5p4",
		oxistates="6, 4*, -2",
		ionenergy=(
				9.0096,
				18.6,
				27.96,
				37.41,
				58.75,
				70.7,
				137.0,
				),
		isotopes={
				105: (104.94364, 0.0),
				106: (105.9375, 0.0),
				107: (106.93501, 0.0),
				108: (107.92944, 0.0),
				109: (108.92742, 0.0),
				110: (109.92241, 0.0),
				111: (110.92111, 0.0),
				112: (111.91701, 0.0),
				113: (112.91589, 0.0),
				114: (113.91209, 0.0),
				115: (114.9119, 0.0),
				116: (115.90846, 0.0),
				117: (116.908645, 0.0),
				118: (117.905828, 0.0),
				119: (118.906404, 0.0),
				120: (119.90402, 0.0009),
				121: (120.904936, 0.0),
				122: (121.9030439, 0.0255),
				123: (122.90427, 0.0089),
				124: (123.9028179, 0.0474),
				125: (124.9044307, 0.0707),
				126: (125.9033117, 0.1884),
				127: (126.9052263, 0.0),
				128: (127.9044631, 0.3174),
				129: (128.9065982, 0.0),
				130: (129.9062244, 0.3408),
				131: (130.9085239, 0.0),
				132: (131.908553, 0.0),
				133: (132.910955, 0.0),
				134: (133.911369, 0.0),
				135: (134.91645, 0.0),
				136: (135.9201, 0.0),
				137: (136.92532, 0.0),
				138: (137.92922, 0.0),
				139: (138.93473, 0.0),
				140: (139.93885, 0.0),
				141: (140.94465, 0.0),
				142: (141.94908, 0.0),
				},
		description=(
				"Silvery metalloid element of group 16. Eight natural isotopes, "
				"nine radioactive isotopes. Used in semiconductors and to a degree "
				"in some steels. Chemistry is similar to Sulphur. Discovered in 1782 "
				"by Franz Miller."
				)
		)

Po = Element(
		84,
		"Po",
		"Polonium",
		group=16,
		period=6,
		block='p',
		series=5,
		mass=208.9824,
		eleneg=2.0,
		eleaffin=1.9,
		covrad=1.46,
		atmrad=1.53,
		vdwrad=0.0,
		tboil=0.0,
		tmelt=527.0,
		density=9.2,
		eleconfig="[Xe] 4f14 5d10 6s2 6p4",
		oxistates="6, 4*, 2",
		ionenergy=(8.414, ),
		isotopes={
				188: (187.999422, 0.0),
				189: (188.998481, 0.0),
				190: (189.995101, 0.0),
				191: (190.994574, 0.0),
				192: (191.991335, 0.0),
				193: (192.99103, 0.0),
				194: (193.988186, 0.0),
				195: (194.98811, 0.0),
				196: (195.985535, 0.0),
				197: (196.98566, 0.0),
				198: (197.983389, 0.0),
				199: (198.983666, 0.0),
				200: (199.981799, 0.0),
				201: (200.98226, 0.0),
				202: (201.980758, 0.0),
				203: (202.98142, 0.0),
				204: (203.980318, 0.0),
				205: (204.981203, 0.0),
				206: (205.980481, 0.0),
				207: (206.981593, 0.0),
				208: (207.9812457, 0.0),
				209: (208.9824304, 0.0),
				210: (209.9828737, 0.0),
				211: (210.9866532, 0.0),
				212: (211.988868, 0.0),
				213: (212.992857, 0.0),
				214: (213.9952014, 0.0),
				215: (214.99942, 0.0),
				216: (216.001915, 0.0),
				217: (217.006335, 0.0),
				218: (218.008973, 0.0),
				219: (219.01374, 0.0),
				220: (220.0166, 0.0),
				},
		description=(
				"Rare radioactive metallic element, belongs to group 16 of the "
				"periodic table. Over 30 known isotopes exist, the most of all "
				"elements. Po-209 has a half-life of 103 years. Possible uses in "
				"heating spacecraft. Discovered by Marie Curie in 1898 in a sample "
				"of pitchblende."
				)
		)

Lv = Element(
		116,
		"Lv",
		"Livermorium",
		group=16,
		period=7,
		block='p',
		series=8,
		isotopes={
				289: (289.19886, 0.0),
				290: (290.19859, 0.0),
				291: (291.20001, 0.0),
				292: (292.19979, 0.0),
				}
		)



#!/usr/bin/env python3
#
#  latex.py
"""
Functions and constants for converting formulae to LaTeX.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Also based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
from typing import Dict, Optional, Sequence, Union

# this package
from ._parser_core import _formula_to_format, _greek_letters

__all__ = ["string_to_latex", "latex_subscript", "latex_superscript"]

_latex_mapping: Dict[str, str] = {k + '-': '\\' + k + '-' for k in _greek_letters}
_latex_mapping["epsilon-"] = "\\varepsilon-"
_latex_mapping["omicron-"] = "o-"
_latex_mapping['.'] = "^\\bullet "
_latex_infix_mapping: Dict[str, str] = {'.': "\\cdot "}


def string_to_latex(
		formula: str,
		prefixes: Optional[Dict[str, str]] = None,
		infixes: Optional[Dict[str, str]] = None,
		suffixes: Sequence[str] = ("(s)", "(l)", "(g)", "(aq)"),
		) -> str:
	r"""
	Convert a formula string to its LaTeX representation.

	:bold-title:`Examples:`

	.. code-block:: python

		>>> string_to_latex('NH4+')
		'NH_{4}^{+}'
		>>> string_to_latex('Fe(CN)6+2')
		'Fe(CN)_{6}^{2+}'
		>>> string_to_latex('Fe(CN)6+2(aq)')
		'Fe(CN)_{6}^{2+}(aq)'
		>>> string_to_latex('.NHO-(aq)')
		'^\bullet NHO^{-}(aq)'
		>>> string_to_latex('alpha-FeOOH(s)')
		'\alpha-FeOOH(s)'

	:param formula: Chemical formula, e.g. ``'H2O'``, ``'Fe+3'``, ``'Cl-'``.
	:param prefixes: Mapping of prefixes to their LaTeX equivalents. Default greek letters and ``.``.
	:no-default prefixes:
	:param infixes: Mapping of infixes to their LaTeX equivalents.
	:param suffixes: Suffixes to keep.

	:return: The LaTeX representation of the formula.
	"""

	if prefixes is None:
		prefixes = _latex_mapping

	if infixes is None:
		infixes = _latex_infix_mapping

	return _formula_to_format(latex_subscript, latex_superscript, formula, prefixes, infixes, suffixes)


def latex_subscript(val: Union[str, float]) -> str:
	"""
	Returns the LaTeX subscript of the given value.

	:param val: The value to superscript
	"""

	return f"_{{{val}}}"


def latex_superscript(val: Union[str, float]) -> str:
	"""
	Returns the LaTeX superscript of the given value.

	:param val: The value to subscript
	"""

	return f"^{{{val}}}"



#!/usr/bin/env python3
#
#  __init__.py
"""
Properties of the chemical elements.

Each chemical element is represented as an object instance. Physicochemical
and descriptive properties of the elements are stored as instance attributes.

Originally created by `Christoph Gohlke <https://www.lfd.uci.edu/~gohlke/>`_.
Licensed under the BSD 3-Clause license


References
----------
1. https://www.nist.gov/pml/atomic-weights-and-isotopic-compositions-relative-atomic-masses
2. https://en.wikipedia.org/wiki/{element.name}


Examples
--------
>>> from chemistry_tools.elements import ELEMENTS
>>> ele = ELEMENTS['C']
>>> ele.number
6
>>> ele.symbol
'C'
>>> ele.name
'Carbon'
>>> ele.description[:21]
'Carbon is a member of'
>>> ele.eleconfig
'[He] 2s2 2p2'
>>> ele.eleconfig_dict
{(1, 's'): 2, (2, 's'): 2, (2, 'p'): 2}
>>> str(ELEMENTS[6])
'Carbon'
>>> len(ELEMENTS)
109
>>> sum(ele.mass for ele in ELEMENTS)
14693.181589001004
>>> for ele in ELEMENTS:
... 	ele.validate()

"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Isotope data from http://www.nist.gov/pml/data/comp.cfm
#

# this package
from ._elements import ELEMENTS, D, H, T
from ._isotope_data import isotope_data  # , undefined_isotopes  # noqa: F401
from ._table import BLOCKS, GROUPS, PERIODS, SERIES
from .actinides import Ac, Am, Bk, Cf, Cm, Es, Fm, Lr, Md, No, Np, Pa, Pu, Th, U
from .alkali_metals import Cs, Fr, K, Li, Na, Rb
from .alkaline_earth_metals import Ba, Be, Ca, Mg, Ra, Sr
from .chalcogens import Lv, O, Po, S, Se, Te
from .classes import Element, Elements, Isotope  # noqa: F401
from .halogens import At, Br, Cl, F, I, Ts
from .lanthanides import Ce, Dy, Er, Eu, Gd, Ho, La, Lu, Nd, Pm, Pr, Sm, Tb, Tm, Yb
from .noble_gases import Ar, He, Kr, Ne, Og, Rn, Xe
from .pnictogens import As, Bi, Mc, N, P, Sb
from .tetrels import C, Fl, Ge, Pb, Si, Sn
from .transition_metals import (
		Ag,
		Au,
		Bh,
		Cd,
		Cn,
		Co,
		Cr,
		Cu,
		Db,
		Ds,
		Fe,
		Hf,
		Hg,
		Hs,
		Ir,
		Mn,
		Mo,
		Mt,
		Nb,
		Ni,
		Os,
		Pd,
		Pt,
		Re,
		Rf,
		Rg,
		Rh,
		Ru,
		Sc,
		Sg,
		Ta,
		Tc,
		Ti,
		V,
		W,
		Y,
		Zn,
		Zr
		)
from .triels import Al, B, Ga, In, Nh, Tl

__all__ = (
		"Element",
		"Isotope",
		"PERIODS",
		"BLOCKS",
		"GROUPS",
		"SERIES",
		"ELEMENTS",
		"period_lengths",
		"accum_period_lengths",  # "groups",
		'H',
		'D',
		'T',
		"Li",  # Alkali Metals
		"Na",
		'K',
		"Rb",
		"Cs",
		"Fr",
		"Be",  # Alkaline Earth Metals
		"Mg",
		"Ca",
		"Sr",
		"Ba",
		"Ra",
		'B',  # Triels
		"Al",
		"Ga",
		"In",
		"Tl",
		"Nh",
		'C',  # Tetrels
		"Si",
		"Ge",
		"Sn",
		"Pb",
		"Fl",
		'N',  # pnictogens
		'P',
		"As",
		"Sb",
		"Bi",
		"Mc",
		'O',  # chalcogens
		'S',
		"Se",
		"Te",
		"Po",
		"Lv",
		'F',  # halogens
		"Cl",
		"Br",
		'I',
		"At",
		"Ts",
		"He",  # noble_gases
		"Ne",
		"Ar",
		"Kr",
		"Xe",
		"Rn",
		"Og",
		"La",  # lanthanides
		"Ce",
		"Pr",
		"Nd",
		"Pm",
		"Sm",
		"Eu",
		"Gd",
		"Tb",
		"Dy",
		"Ho",
		"Er",
		"Tm",
		"Yb",
		"Lu",
		"Ac",  # actinides
		"Th",
		"Pa",
		'U',
		"Np",
		"Pu",
		"Am",
		"Cm",
		"Bk",
		"Cf",
		"Es",
		"Fm",
		"Md",
		"No",
		"Lr",
		"Sc",  # transition_metals
		"Ti",
		'V',
		"Cr",
		"Mn",
		"Fe",
		"Co",
		"Ni",
		"Cu",
		"Zn",
		'Y',
		"Zr",
		"Nb",
		"Mo",
		"Tc",
		"Ru",
		"Rh",
		"Pd",
		"Ag",
		"Cd",
		"Hf",
		"Ta",
		'W',
		"Re",
		"Os",
		"Ir",
		"Pt",
		"Au",
		"Hg",
		"Rf",
		"Db",
		"Sg",
		"Bh",
		"Hs",
		"Mt",
		"Ds",
		"Rg",
		"Cn",
		)

period_lengths = (2, 8, 8, 18, 18, 32, 32)
accum_period_lengths = (2, 10, 18, 36, 54, 86, 118)

# icosagens, crystallogens, pnictogens, chalcogens, halogens
groups = {g: tuple(x - 18 + g for x in accum_period_lengths[1:]) for g in range(13, 18)}
groups[1] = (1, ) + tuple(x + 1 for x in accum_period_lengths[:-1])  # alkali metals
groups[2] = tuple(x + 2 for x in accum_period_lengths[:-1])  # alkaline earth metals
groups[18] = accum_period_lengths  # noble gases

if __name__ == "__main__":
	# stdlib
	import doctest

	print(f"ELEMENTS = {repr(ELEMENTS)}")
	doctest.testmod(verbose=False)



#!/usr/bin/env python3
#
#  _table.py
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#

# stdlib
from typing import Dict, Tuple

__all__ = ("PERIODS", "BLOCKS", "GROUPS", "SERIES")

PERIODS: Dict[int, str] = {1: 'K', 2: 'L', 3: 'M', 4: 'N', 5: 'O', 6: 'P', 7: 'Q'}

BLOCKS = {'s': '', 'g': '', 'f': '', 'd': '', 'p': ''}

GROUPS: Dict[int, Tuple[str, str]] = {
		1: ("IA", "Alkali metals"),
		2: ("IIA", "Alkaline earths"),
		3: ("IIIB", ''),
		4: ("IVB", ''),
		5: ("VB", ''),
		6: ("VIB", ''),
		7: ("VIIB", ''),
		8: ("VIIIB", ''),
		9: ("VIIIB", ''),
		10: ("VIIIB", ''),
		11: ("IB", "Coinage metals"),
		12: ("IIB", ''),
		13: ("IIIA", "Boron group"),
		14: ("IVA", "Carbon group"),
		15: ("VA", "Pnictogens"),
		16: ("VIA", "Chalcogens"),
		17: ("VIIA", "Halogens"),
		18: ("VIIIA", "Noble gases"),
		}

SERIES: Dict[int, str] = {
		1: "Nonmetals",
		2: "Noble gases",
		3: "Alkali metals",
		4: "Alkaline earth metals",
		5: "Metalloids",
		6: "Halogens",
		7: "Poor metals",
		8: "Transition metals",
		9: "Lanthanides",
		10: "Actinides",
		}



#!/usr/bin/env python3
#
#  enums.pyi
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from typing import Any

# 3rd party
from enum_tools import IntEnum, StrEnum

class PubChemNamespace(StrEnum):
	CID = Cid = cid = "cid"
	Name = NAME = name = "name"
	SMILES = Smiles = smiles = "smiles"
	INCHIKEY = Inchikey = inchikey = "inchikey"

	def __new__(cls, value: str, doc: str): ...

	@classmethod
	def is_valid_value(cls, value: Any) -> bool: ...

# @document_enum

class PubChemFormats(StrEnum):
	"""
	Enum of supported formats for the PubChem REST API.
	"""
	JSON = Json = json = "JSON"
	XML = Xml = xml = "XML"
	CSV = csv = Csv = "CSV"
	PNG = png = Png = "PNG"

	def __new__(cls, value: str): ...

	@classmethod
	def is_valid_value(cls, value: Any) -> bool: ...

class CoordinateType(IntEnum):
	TWO_D = ...
	THREE_D = ...
	SUBMITTED = ...
	EXPERIMENTAL = ...
	COMPUTED = ...
	STANDARDIZED = ...
	AUGMENTED = ...
	ALIGNED = ...
	COMPACT = ...
	UNITS_ANGSTROMS = ...
	UNITS_NANOMETERS = ...
	UNITS_PIXEL = ...
	UNITS_POINTS = ...
	UNITS_STDBONDS = ...
	UNITS_UNKNOWN = ...

	@classmethod
	def is_valid_value(cls, value: Any) -> bool: ...



#!/usr/bin/env python3
#
#  enums.py
"""
Enumerations.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from typing import Any

# 3rd party
from enum_tools import IntEnum, StrEnum

__all__ = ["PubChemNamespace", "PubChemFormats", "CoordinateType"]


class PubChemNamespace(StrEnum):
	"""
	Enumeration of possible values for the PubChem namespace.
	"""

	CID = Cid = cid = "cid", "PubChem Compound ID"
	Name = NAME = name = "name", "Compound Name"
	SMILES = Smiles = smiles = "smiles", "SMILES String"
	INCHIKEY = Inchikey = inchikey = "inchikey", "InChI Key"

	# INCHI = Inchi = inchi = "inchi"# TODO: requires argument
	# Formula = FORMULA = formula = "formula"
	# SDF = Sdf = sdf = "sdf"

	# TODO: listkey for formula lookup https://pubchemdocs.ncbi.nlm.nih.gov/pug-rest$_Toc494865583

	def __new__(cls, value: str, doc: str):  # noqa: D102
		obj = str.__new__(cls, value)  # noqa
		obj._value_ = value
		obj.__doc__ = doc
		return obj

	@classmethod
	def is_valid_value(cls, value: Any) -> bool:
		"""
		Returns whether the value is a valid member of this :class:`enum.Enum`.

		:param value:
		"""

		return str(value) in {str(item) for item in PubChemNamespace}


# @document_enum
class PubChemFormats(StrEnum):
	"""
	Enumeration of supported formats for the PubChem REST API.
	"""

	JSON = Json = json = "JSON"
	XML = Xml = xml = "XML"
	CSV = csv = Csv = "CSV"
	PNG = png = Png = "PNG"

	def __new__(cls, value: str):  # noqa: D102
		obj = str.__new__(cls, value)  # noqa
		obj._value_ = value
		obj.__doc__ = f"{value} Format"
		return obj

	@classmethod
	def is_valid_value(cls, value: Any) -> bool:
		"""
		Returns whether the value is a valid member of this :class:`enum.Enum`.

		:param value:
		"""

		return str(value).upper() in {str(item) for item in PubChemFormats}


class CoordinateType(IntEnum):
	"""
	Enumeration of valid values for the coordinate type.
	"""

	TWO_D = 1
	THREE_D = 2
	SUBMITTED = 3
	EXPERIMENTAL = 4
	COMPUTED = 5
	STANDARDIZED = 6
	AUGMENTED = 7
	ALIGNED = 8
	COMPACT = 9
	UNITS_ANGSTROMS = 10
	UNITS_NANOMETERS = 11
	UNITS_PIXEL = 12
	UNITS_POINTS = 13
	UNITS_STDBONDS = 14
	UNITS_UNKNOWN = 255

	@classmethod
	def is_valid_value(cls, value: Any) -> bool:
		"""
		Returns whether the value is a valid member of this :class:`enum.Enum`.

		:param value:
		"""

		return str(value).upper() in {str(item) for item in CoordinateType}
